"""Utility functions for the project."""

import collections
import contextlib
import signal
import subprocess
import time
import types

from . import cliargs


def debug(message: str) -> None:
    """Output debug message if debug mode is enabled."""
    if cliargs.OPTIONS.debug:
        print(message)


class AlarmError(Exception):
    """Exception thrown on timeout."""


def _alarm_handler(_signum: int, _frame: types.FrameType | None) -> None:
    """Raise an AlarmError."""
    raise AlarmError()


@contextlib.contextmanager
def timeout(timeout_s: int) -> collections.abc.Iterator[None]:
    """Use signal.alarm to trigger a timeout after some seconds.

    An AlarmError will be raised if the timeout is hit.
    """
    signal.signal(signal.SIGALRM, _alarm_handler)
    signal.alarm(timeout_s)
    try:
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)


def run_once(args: list[str], *, timeout_s: int = 600) -> subprocess.CompletedProcess[str]:
    """Run a command and return the completed process.

    On timeout, changes the return code to 127.
    """
    try:
        return subprocess.run(
            args,
            encoding="utf8",
            capture_output=True,
            timeout=timeout_s,
            check=False,
        )
    except subprocess.TimeoutExpired as e:
        return subprocess.CompletedProcess(
            args,
            127,
            e.stdout.decode("utf8") if e.stdout else "",
            e.stderr.decode("utf8") if e.stderr else "",
        )


def run_repeatedly(
    args: list[str],
    *,
    timeout_s: int = 600,
    max_attempts: int = 5,
    pause_s: int = 60,
) -> subprocess.CompletedProcess[str]:
    """Run command until it exits with 0 or we run out of retries.

    Returns the last completed process.

    On timeout, changes the return code to 127.
    """
    for i in range(max_attempts):
        time.sleep(i * pause_s)
        p = run_once(args, timeout_s=timeout_s)
        if p.returncode == 0:
            break
        print(f"Command {args} failed (try #{i + 1}):")
        print(p.stderr)
    return p
